from django.conf import settings


METADATA_KEY = getattr(settings, 'METADATA_KEY', 'metadata:%(identifier)s:%(id)s')

REDIS_CONNECTION_CLASS = getattr(settings, 'METADATA_REDIS_CONNECTION_CLASS', None)

REDIS_CONNECTION = getattr(settings, 'METADATA_REDIS_CONNECTION', {})



from django.db import models

from metadata.mixins import MetadataMixin


class Poll(MetadataMixin, models.Model):
    question = models.CharField(max_length=200)
    pub_date = models.DateTimeField('date published')

    def __str__(self):
        return self.question

    def refresh(self):
        return self.__class__._default_manager.get(pk=self.pk)



from .models import MetadataContainer

from . import settings
from .connection import client


class MetadataMixin(object):
    metadata = MetadataContainer(connection=client,
                                 key=lambda instance: instance.metadata_key)

    @property
    def metadata_key(self):
        key = getattr(self, 'METADATA_KEY', settings.METADATA_KEY)

        if key:
            return key % {
                'identifier': self.__class__.__name__.lower(),
                'id': self.pk
            }

        return None



from django.test import TestCase
from django.utils import timezone as datetime

from exam import fixture, before, Exam

from metadata.connection import client

from .models import Poll


class MetadataTest(Exam, TestCase):
    @before
    def flush_database(self):
        client.flushdb()

    @fixture
    def poll(self):
        return Poll.objects.create(question='What\'s your favorite color', pub_date=datetime.now())

    def test_basic_storing(self):
        self.poll.metadata['key'] = 'value'

        poll = self.poll.refresh()

        assert poll.metadata['key'] == 'value'

        self.poll.metadata = {
            'key': 'valeur'
        }

        poll = self.poll.refresh()

        assert poll.metadata['key'] == 'valeur'

    def test_get_or_set(self):
        assert self.poll.metadata.get_or_set('key', lambda: 'value') == 'value'

        assert self.poll.metadata['key'] == 'value'

        assert self.poll.metadata.get_or_set('key', lambda: 'test') == 'value'

    def test_delete_key(self):
        self.poll.metadata['key'] = 'value'

        self.poll.metadata['key'] = None

        poll = self.poll.refresh()

        self.assertNotIn('key', poll.metadata)

        poll.metadata['key'] = 'value'

        poll = self.poll.refresh()

        self.assertIn('key', poll.metadata)

        poll.metadata = {
            'key': None
        }

        poll = self.poll.refresh()

        self.assertNotIn('key', poll.metadata)

        with self.assertRaises(KeyError):
            poll.metadata['key']

        assert poll.metadata.get('key', None) is None

        poll.metadata = {
            'key': None
        }

        del poll.metadata['key']

        with self.assertRaises(KeyError):
            poll.metadata['key']

        assert poll.metadata.get('key', None) is None

    def test_wildcard_delete(self):
        self.poll.metadata['key_1'] = 'value'
        self.poll.metadata['key_2'] = 'value'
        self.poll.metadata['key_3'] = 'value'
        self.poll.metadata['diff_key_3'] = 'value'

        del self.poll.metadata['key_*']

        poll = self.poll.refresh()

        values = poll.metadata.keys()

        assert values == ['diff_key_3']

    def test_iteration(self):
        keys = {
            'key1': 'value',
            'key2': 'value',
        }

        self.poll.metadata = keys

        poll = self.poll.refresh()

        for k, v in poll.metadata.items():
            assert keys[k] == v

    def test_incr(self):
        self.poll.metadata.incr('key', 2)

        poll = self.poll.refresh()

        assert int(poll.metadata['key']) == 2

    def test_values(self):
        keys = {
            'key1': 'value1',
            'key2': 'value2',
        }

        self.poll.metadata = keys

        poll = self.poll.refresh()

        self.assertIn('value2', poll.metadata.values())
        self.assertIn('value1', poll.metadata.values())

    def test_keys(self):
        keys = {
            'key1': 'value1',
            'key2': 'value2',
        }

        self.poll.metadata = keys

        poll = self.poll.refresh()

        self.assertIn('key2', poll.metadata.keys())
        self.assertIn('key1', poll.metadata.keys())



import six

from fnmatch import fnmatch


class MetadataContainer(object):
    def __init__(self, connection, key, instance=None):
        self._connection = connection
        self._key = key
        self.loaded = False
        self.instance = instance

    @property
    def metadata(self):
        if not self.loaded:
            self.reload()

        return self._metadata

    @property
    def key(self):
        if callable(self._key):
            return self._key(self.instance)

        return self._key

    @property
    def connection(self):
        if callable(self._connection):
            return self._connection(self.instance)

        return self._connection

    def __get__(self, instance, owner=None):
        if instance:
            return self.copy(instance=instance)

        return self

    def __set__(self, instance, metadata):
        to_update = dict((key, value)
                         for key, value in six.iteritems(metadata)
                         if value is not None)

        to_delete = [k for k, v in six.iteritems(metadata)
                     if v is None]

        container = self.__get__(instance)

        redis = container.connection

        if to_delete:
            redis = redis.pipeline()

        if to_update:
            redis.hmset(container.key, to_update)

        if to_delete:
            for key in to_delete:
                redis.hdel(container.key, key)

            redis.execute()

        container.reload()

        return container

    def get_or_set(self, key, func):
        value = self.get(key)

        if value is None:
            value = func()

            self[key] = value

        return value

    def reload(self):
        self._metadata = self.connection.hgetall(self.key) or {}

    def incr(self, key, incrby=1):
        return self.connection.hincrby(self.key, key, incrby)

    def listkeys(self, key):
        keys = [key, ]

        if '*' in key:
            keys = [result
                    for result in self.connection.hkeys(self.key)
                    if fnmatch(result, key)]

        return keys

    def __delitem__(self, key):
        with self.connection.pipeline() as pipe:
            for key in self.listkeys(key):
                pipe.hdel(self.key, key)

            pipe.execute()

    def __getitem__(self, key):
        if key in self.metadata:
            return self.metadata[key]

        raise KeyError

    def __setitem__(self, key, value=None):
        if value is None:
            self.connection.hdel(self.key, key)
        else:
            self.connection.hset(self.key, key, value)

    def __copy__(self, instance=None):
        return self.__class__(self._connection,
                              self._key,
                              instance=instance or self.instance)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def iteritems(self):
        for item in self.metadata.items():
            yield item

    def iterkeys(self):
        for k, v in self.iteritems():
            yield k

    def itervalues(self):
        for k, v in self.iteritems():
            yield v

    def items(self):
        return list(self.iteritems())

    def keys(self):
        return list(self.iterkeys())

    def values(self):
        return list(self.itervalues())

    def has_key(self, key):
        return key in self.metadata

    __contains__ = has_key
    __iter__ = iterkeys

    def copy(self, instance=None):
        return self.__copy__(instance=instance)

    def __str__(self):
        return str(dict(self.items()))
